// Rouslan Synth 2011-2015
#include "declarations.h"
#include "tables.h"
#include "synth.h"
#include "midi.h"
#include "ui.h"
#include "lcd.h"

#if SIMULATOR
extern u8 eeprom[128];
#define save(v) printf("%d ", (u32)v);
#define load(v) v = eeprom[addr++];
#else
#include <avr/eeprom.h>

#define save(v) eeprom_update_byte((u8*)addr++, v);
#define load(v) v = eeprom_read_byte((u8*)addr++);
#endif

channel chan[N_CHAN];

u16 addr;
u8	cur_patch;

void synth_save() // 32B in EEPROM
{
	u8 i,j;
	channel *c = &chan[0];
	addr = 32*cur_patch;

	for(i=0; i<2; i++) {
		oscillator *o = &c->osc[i];
		save(o->type);
		save(o->tweak);
		save(o->note_offset);
		
		for(j=0; j<ENV_POINTS; j++) {
			save(o->env_vol.pts[j].value);
			save(o->env_vol.pts[j].rate);
		}
	}

	save(c->ring);
	save(c->slide_rate);
	for(j=0; j<ENV_POINTS; j++) {
		save(c->env_low_pass.pts[j].value);
		save(c->env_low_pass.pts[j].rate);
	}
}

void synth_load()
{
	u8 i,j;
	channel *c = &chan[0];
	addr = 32*cur_patch;

	for(i=0; i<2; i++) {
		oscillator *o = &c->osc[i];
		load(o->type);
		load(o->tweak);
		load(o->note_offset);
		
		for(j=0; j<ENV_POINTS; j++) {
			load(o->env_vol.pts[j].value);
			load(o->env_vol.pts[j].rate);
		}
	}

	load(c->ring);
	load(c->slide_rate);
	for(j=0; j<ENV_POINTS; j++) {
		load(c->env_low_pass.pts[j].value);
		load(c->env_low_pass.pts[j].rate);
	}
}

void synth_init()
{
	u8 i;
	for(i=0; i<N_CHAN; i++) {
		chan[i].state = OFF;
	}
	synth_load();
}

inline void synth_env_update(envelope *e)
{
	s8 rate;

	if(e->point == ENV_POINTS - 1)
		return;

	rate = e->pts[e->point].rate;

	if(rate > 0 && e->value > e->pts[e->point + 1].value - rate ||
	   rate < 0 && e->value + rate < e->pts[e->point + 1].value) {
		e->point++;
		e->value = e->pts[e->point].value;
	} else
		e->value += rate;
}

void compute_ds(channel* c, u8 set_cur_ds)
{
	u8 i;
	u8 note = c->notes[c->note_idx];

	for(i=0; i<2; i++) {
		u8 osc_note = note + c->osc[i].note_offset;
		u16	ds		= note2offset[osc_note];
		u16 up_ds	= note2offset[osc_note + 2];

		ds += ((s32)(up_ds - ds) * c->bend_modulation >> 13);

		c->osc[i].target_ds = ds;
		if(set_cur_ds)
			c->osc[i].ds = ds;
	}
}

inline s8 full_period_sine(u8 idx)
{
	s8 s = sine[idx & WAVET_SIZE - 1];
	return idx >= WAVET_SIZE ? -s : s;
}

void synth_update()
{
	channel *c = &chan[0];

	if(c->state != OFF) {
		u8 i;
		for(i=0; i<2; i++) {
			// mono slide
			s16 dif = (s16)c->osc[i].target_ds - c->osc[i].ds;
			s8 sign = dif > 0 ? 1 : (dif < 0 ? -1 : 0);
			c->osc[i].ds += sign + (dif >> 8 - c->slide_rate);
		}
		for(i=0; i<2; i++)
			synth_env_update(&c->osc[i].env_vol);

		synth_env_update(&c->env_low_pass);

		c->inst_low_pass = c->env_low_pass.value + c->breadth;
		if(c->inst_low_pass < 5)
			c->inst_low_pass = 5;

		if(c->mod != 0xFF) {
			c->bend_modulation = c->bend + (c->mod * full_period_sine(4*ctrl_cnt & (WAVET_SIZE - 1)) >> 2);
			compute_ds(c, 1);
			if(c->mod == 0)
				c->mod = 0xFF;
		}

		if(c->state == DECAYING) {
			if(c->vol < 5) {
				c->state = OFF;
				c->vol = 0;
			} else
				c->vol -= 5;
		}
	}
}

u8 ctrl_bank;
u8 cur_ctrl;

void synth_update_controls(u8 val)
{
	channel *c = &chan[0];

	switch(ctrl_bank) {
		case CTRL_OSC_SHAPE:
		{
			switch(cur_ctrl) {
				case 0:	c->osc[0].type = val / 64; break;
				case 1:	c->osc[0].tweak = val; break;
				case 2:	c->osc[1].type = val / 64; break;
				case 3:	c->osc[1].tweak = val; break;
			}
			break;
		}
		case CTRL_OSC0_ENV_VOL_VALUE:
		case CTRL_OSC1_ENV_VOL_VALUE:
			c->osc[(ctrl_bank - CTRL_OSC0_ENV_VOL_VALUE) >> 1].env_vol.pts[cur_ctrl].value = val;
			break;
		case CTRL_OSC0_ENV_VOL_RATE:
		case CTRL_OSC1_ENV_VOL_RATE:
			c->osc[(ctrl_bank - CTRL_OSC0_ENV_VOL_VALUE) >> 1].env_vol.pts[cur_ctrl].rate = (s8)(val >> 1) - 64;
			break;
		case CTRL_LP_ENV_VOL_VALUE:
			c->env_low_pass.pts[cur_ctrl].value = val;
			break;
		case CTRL_LP_ENV_VOL_RATE:
			c->env_low_pass.pts[cur_ctrl].rate = (s8)(val >> 1) - 64;
			break;
		case CTRL_TRANSPOSE_AND_RING_MOD:
			switch(cur_ctrl) {
				case 0:
				case 1: c->osc[cur_ctrl].note_offset = val/4 - 32; break;
				case 2: c->slide_rate = val / 32; break;
				case 3: c->ring = val / 32; break;
			}
			break;
		case CTRL_PATCH_SAVE:
		case CTRL_PATCH_RESTORE:
			if(cur_ctrl == 0) cur_patch = val/64;
			break;
	}
	menu_update_val(val);
}

u8  ctrl_counter;
u8	lcd_counter;
u32 noise = 1;

#define BUF_SIZE 64
u8  output[BUF_SIZE];
u8	wr_idx = 1, rd_idx;

void synth_sample()
{
	while(wr_idx != rd_idx) {
		u8 i, j;

		// update state
		if(++lcd_counter >= SMPLS_PER_LCD) {
			lcd_update();
			lcd_counter = 0;
			if(++ctrl_counter >= LCD_PER_CTRL) {
				control();
				ctrl_counter = 0;
			}
		}

		// generate wave
		for(i=0;i<1;i++) {
			channel *c = &chan[i];
			s8 osc[2];

			for(j=0;j<2;j++) {
				oscillator *o = &c->osc[j];

				u16 smp_idx = o->cur_sample;

				switch(o->type) {
					case SAWTOOTH:
					{
						u8 val = smp_idx >> 8;
						osc[j] = -128 + val + (val * o->tweak >> 6);
						break;
					}
					case SQUARE:
						osc[j] = smp_idx >> 8 > o->tweak ? -128:127;
						break;
					case SINE:
					{
						u8 wave_idx;
						wave_idx = (smp_idx >> 10) & (WAVET_SIZE-1);
						osc[j] = full_period_sine(wave_idx) + o->tweak;
						break;
					}
					case NOISE:
					{ 
						static u8 prev_smp;
						if(prev_smp != smp_idx >> 8) {
							prev_smp = smp_idx >> 8;
							noise = (noise << 1) | (noise ^ (noise >> 13)) & 1;
						}
						osc[j] = noise & 0xFF;
					}
				}
				osc[j] = osc[j] * o->env_vol.value >> 8;
				o->cur_sample += o->ds;
			}

			if(!c->ring)
				c->output = (osc[0] + osc[1] >> 1) * c->vol;
			else
				c->output = (osc[0] * osc[1] >> (8 - c->ring)) * c->vol;

			// low pass filtering
			c->output = c->last_output + (u16)((u32)(c->output - c->last_output) * (u8)c->inst_low_pass >> 8); 
			c->last_output = c->output;
		}

		output[wr_idx] = 128 + (s8)(chan[0].output >> 8);
		wr_idx = wr_idx + 1 & BUF_SIZE - 1;
	}
}

#if SIMULATOR
u8 TIMER0_COMPA_vect()
#else
ISR(TIMER0_COMPA_vect) // Timer interrupt
#endif
{
	u8 r = output[rd_idx];
#if SIMULATOR
	if((rd_idx + 1 & BUF_SIZE - 1) == wr_idx)
		synth_sample();
	rd_idx = rd_idx + 1 & BUF_SIZE - 1;
	return r;
#else
	rd_idx = rd_idx + 1 & BUF_SIZE - 1;
	OCR1A = OCR1B = r;
#endif
}
